=== modified file 'baz_import.py'
--- baz_import.py	2005-10-23 22:56:20 +0000
+++ baz_import.py	2005-10-27 21:12:45 +0000
@@ -14,8 +14,12 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-from bzrlib.errors import NotBranchError, BzrCommandError
+import errno
+
+from bzrlib.errors import BzrError
+from bzrlib.errors import NotBranchError, BzrCommandError, NoSuchRevision
 from bzrlib.branch import Branch
+from bzrlib.clone import copy_branch
 from bzrlib.commit import Commit, NullCommitReporter
 from bzrlib.commands import Command
 from bzrlib.option import _global_option
@@ -33,7 +37,6 @@
 import os.path
 import shutil
 import bzrlib
-from bzrlib.errors import BzrError
 import bzrlib.trace
 import bzrlib.merge
 import bzrlib.inventory
@@ -236,11 +239,35 @@
             "Directory \"%s\" already exists, and the last revision is not"
             " an Arch revision (%s)" % (branch.base, last_patch))
 
+def do_branch(br_from, to_location, revision_id):
+    """Derived from branch in builtins."""
+    br_from.lock_read()
+    try:
+        try:
+            os.mkdir(to_location)
+        except OSError, e:
+            if e.errno == errno.EEXIST:
+                raise UserError('Target directory "%s" already'
+                                      ' exists.' % to_location)
+            if e.errno == errno.ENOENT:
+                raise UserError('Parent of "%s" does not exist.' %
+                                      to_location)
+            else:
+                raise
+        try:
+            copy_branch(br_from, to_location, revision_id, None)
+        except NoSuchRevision:
+            rmtree(to_location)
+            msg = "The branch %s has no revision %s." % (from_location, revision_id)
+            raise UserError(msg)
+    finally:
+        br_from.unlock()
 
-def get_remaining_revisions(output_dir, version):
+def get_remaining_revisions(output_dir, version, reuse_history_from=[]):
     last_patch = None
     old_revno = None
-    if os.path.exists(output_dir):
+    output_exists = os.path.exists(output_dir)
+    if output_exists:
         # We are starting from an existing directory, figure out what
         # the current version is
         branch = Branch.open(output_dir)
@@ -252,6 +279,27 @@
 
     try:
         ancestors = version_ancestry(version)
+        if not output_exists and reuse_history_from != []:
+            for ancestor in reversed(ancestors):
+                if last_patch is not None:
+                    # found something to copy
+                    break
+                # try to grab a copy of ancestor
+                # note that is not optimised: we could look for namespace
+                # transitions and only look for the past after the 
+                # transition.
+                for history_root in reuse_history_from:
+                    possible_source = os.path.join(history_root,
+                        map_namespace(ancestor.version))
+                    try:
+                        source = Branch.open(possible_source)
+                        rev_id = revision_id(ancestor)
+                        if rev_id in source.revision_history():
+                            do_branch(source, output_dir, rev_id)
+                            last_patch = ancestor
+                            break
+                    except NotBranchError:
+                        pass
     except NoSuchVersion, e:
         raise UserError(str(e))
 
@@ -270,8 +318,25 @@
         ancestors = ancestors[i+1:]
     return ancestors, old_revno
 
+
+###class Importer(object):
+###    """An importer.
+###    
+###    Currently this is used as a parameter object, though more behaviour is
+###    possible later.
+###    """
+###
+###    def __init__(self, output_dir, version, printer, fancy=True, fast=False,
+###                 verbose=False, dry_run=False, max_count=None, 
+###                   reuse_history_from=[]):
+###        self.output_dir = output_dir
+###        self.version = version
+###        self.
+
+
 def import_version(output_dir, version, printer, fancy=True, fast=False,
-                   verbose=False, dry_run=False, max_count=None):
+                   verbose=False, dry_run=False, max_count=None,
+                   reuse_history_from=[]):
     """
     >>> q = test_environ()
     >>> result_path = os.path.join(q, "result")
@@ -306,7 +371,8 @@
     >>> teardown_environ(q)
     """
     try:
-        ancestors, old_revno = get_remaining_revisions(output_dir, version)
+        ancestors, old_revno = get_remaining_revisions(output_dir, version,
+                                                       reuse_history_from)
     except NotBranchError, e:
         raise UserError("%s exists, but is not a bzr branch." % output_dir)
     if old_revno is None and len(ancestors) == 0:
@@ -504,7 +570,7 @@
         log_description = log.description
         is_continuation = log.continuation_of is not None
         log_creator = log.creator
-        direct_merges = get_direct_merges(revdir, revision)
+        direct_merges = get_direct_merges(revdir, revision, log)
 
         timestamp = email.Utils.mktime_tz(log_date + (0,))
         if log_summary is None:
@@ -520,8 +586,8 @@
                 # if we want it to be in revision-history, do that here.
                 branch.add_pending_merge(revision_id(missing_ancestor))
                 missing_ancestor = None
-            for merge in direct_merges:
-                branch.add_pending_merge(revision_id(merge.revision))
+            for merged_rev in direct_merges:
+                branch.add_pending_merge(revision_id(merged_rev))
             branch.set_inventory(baz_inv)
             commitobj = Commit(reporter=ImportCommitReporter(pb))
             commitobj.commit(branch, log_message.decode('ascii', 'replace'), 
@@ -532,16 +598,18 @@
     yield Progress("revisions", len(ancestors), len(ancestors))
     unlink_unversioned(branch, revdir)
 
-def get_direct_merges(revdir, revision):
-    if pybaz.WorkingTree(revdir).tree_version != revision.version:
-        pybaz.WorkingTree(revdir).set_tree_version(revision.version)
+def get_direct_merges(revdir, revision, log):
+    continuation = log.continuation_of
+    previous_version = revision.version
+    if pybaz.WorkingTree(revdir).tree_version != previous_version:
+        pybaz.WorkingTree(revdir).set_tree_version(previous_version)
     log_path = "%s/{arch}/%s/%s/%s/%s/patch-log/%s" % (revdir, 
         revision.category.nonarch, revision.branch.nonarch, 
         revision.version.nonarch, revision.archive, revision.patchlevel)
     temp_path = tempfile.mktemp(dir=os.path.dirname(revdir))
     os.rename(log_path, temp_path)
     merges = list(iter_new_merges(revdir, revision.version))
-    direct = direct_merges (merges)
+    direct = direct_merges(merges, [continuation])
     os.rename(temp_path, log_path)
     return direct
 
@@ -611,14 +679,14 @@
 _global_option('max-count', type = int)
 class cmd_baz_import_branch(Command):
     """Import an Arch or Baz branch into a bzr branch"""
-    takes_args = ['to_location', 'from_branch?']
+    takes_args = ['to_location', 'from_branch?', 'reuse_history*']
     takes_options = ['verbose', 'max-count']
 
     def printer(self, name):
         print name
 
     def run(self, to_location, from_branch=None, fast=False, max_count=None,
-            verbose=False, dry_run=False):
+            verbose=False, dry_run=False, reuse_history_list=[]):
         to_location = os.path.realpath(str(to_location))
         if from_branch is not None:
             try:
@@ -626,33 +694,50 @@
             except pybaz.errors.NamespaceError:
                 print "%s is not a valid Arch branch." % from_branch
                 return 1
+        if reuse_history_list is None:
+            reuse_history_list = []
         import_version(to_location, from_branch, self.printer, 
-                       max_count=max_count)
+                       max_count=max_count, reuse_history_from=reuse_history_list)
 
 
 class cmd_baz_import(Command):
-    """Import an Arch or Baz archive into bzr branches."""
-    takes_args = ['to_root_dir', 'from_archive']
+    """Import an Arch or Baz archive into bzr branches.
+    
+    reuse_history allows you to specify any previous imports you 
+    have done of different archives, which this archive has branches
+    tagged from. This will dramatically reduce the time to convert 
+    the archive as it will not have to convert the history already
+    converted in that other branch.
+    """
+    takes_args = ['to_root_dir', 'from_archive', 'reuse_history*']
     takes_options = ['verbose']
 
     def printer(self, name):
         print name
 
-    def run(self, to_root_dir, from_archive, verbose=False):
+    def run(self, to_root_dir, from_archive, verbose=False,
+            reuse_history_list=[]):
+        if reuse_history_list is None:
+            reuse_history_list = []
         to_root = str(os.path.realpath(to_root_dir))
         if not os.path.exists(to_root):
             os.mkdir(to_root)
-        import_archive(to_root, from_archive, verbose, self.printer)
-
-
-def import_archive(to_root, from_archive, verbose, printer):
+        import_archive(to_root, from_archive, verbose, self.printer, 
+                       reuse_history_list)
+
+
+def import_archive(to_root, from_archive, verbose, printer,
+                   reuse_history_from=[]):
+    real_to = os.path.realpath(to_root)
+    history_locations = [real_to] + reuse_history_from
     for version in pybaz.Archive(str(from_archive)).iter_versions():
         target = os.path.join(to_root, map_namespace(version))
         printer("importing %s into %s" % (version, target))
         if not os.path.exists(os.path.dirname(target)):
             os.makedirs(os.path.dirname(target))
         try:
-            import_version(target, version, printer)
+            import_version(target, version, printer,
+                           reuse_history_from=reuse_history_from)
         except pybaz.errors.ExecProblem,e:
             if str(e).find('The requested revision cannot be built.') != -1:
                 printer("Skipping version %s as it cannot be built due"

=== modified file 'fai.py'
--- fai.py	2005-10-23 00:10:38 +0000
+++ fai.py	2005-10-24 07:01:08 +0000
@@ -22,7 +22,7 @@
 __docformat__ = "restructuredtext"
 __doc__ = "Utility functions to be used by commands"
 
-def direct_merges(merges):
+def direct_merges(merges, excludes=[]):
     """Get a list of direct merges, from a list of direct and indirect
     
     :param merges: Iterator of merge patchlogs
@@ -30,20 +30,23 @@
     :return: The direct merges
     :rtype: list of `pybaz.Patchlog`
     """
-    indirect = []
-    direct = []
+    indirect = set()
     logs = list(merges)
     if not logs:
         return []
     for log in logs:
         try:
-            this_indirect = [f for f in log.new_patches if f != log.revision]
+            this_indirect = set([str(f) for f in log.new_patches 
+                                 if f != log.revision and 
+                                     str(f) not in indirect])
         except pybaz.errors.NamespaceError:
             print
             print "log ", log, " unusable, attempting to use archive copy."
             log = pybaz.Revision(str(log.revision)).patchlog
-            this_indirect = [f for f in log.new_patches if f != log.revision]
-        indirect.extend(this_indirect)
+            this_indirect = set([str(f) for f in log.new_patches 
+                                 if f != log.revision and 
+                                     str(f) not in indirect])
+        indirect.update(this_indirect)
         if log.continuation_of is not None:
             # continuations list everything in new_patches
             continue
@@ -68,8 +71,10 @@
                         found = True
                 print "ancestor of %s is %s" % (log.revision, ancestor)
         if ancestor is not None:
-            indirect.append(ancestor)
-    return [log for log in logs if not log.revision in indirect]
+            indirect.add(str(ancestor))
+    return [log.revision for log in logs if not str(log.revision) in indirect
+            and log.revision not in excludes]
+
 
 def namespace_previous(revision):
     if revision.patchlevel == 'base-0':

=== modified file 'tests/test_baz_import.py'
--- tests/test_baz_import.py	2005-10-23 22:56:20 +0000
+++ tests/test_baz_import.py	2005-10-24 07:27:27 +0000
@@ -157,6 +157,19 @@
         # tag into the kept archive
         pybaz.Revision(self._missing_import + '--base-0').make_continuation(
             pybaz.Version(self._missing_ancestor))
+
+        # make an import for testing history-reuse logic.
+        def collect(text):
+            pass
+        # note the use of a namespace layout here.
+        self._missing_import_imported = os.path.join(self._tmpdir, 
+                                                     'archivegone-bzr')
+        os.mkdir(os.path.join(self._tmpdir, 'archivegone-bzr'))
+        os.mkdir(os.path.join(self._tmpdir, 'archivegone-bzr', 'c'))
+        import_version(os.path.join(self._tmpdir, 'archivegone-bzr', 
+                                    'c', 'import'),
+                       pybaz.Version(self._missing_import),
+                       collect)
         # and make it inaccessible
         pybaz.Archive('demo-gone@DONOTUSE').unregister()
 
@@ -271,6 +284,8 @@
         self.assertEqual(rev.committer, "Test User<test@example.org>")
         self.assertEqual(rev.message, "tag of demo@DONOTUSE/c--import--0--base-0")
         self.assertEqual(rev.revision_id, self._baz._empty_merged_tag_bzr_base)
+        self.assertEqual(['Arch-1:demo@DONOTUSE%c--import--0--base-0'],
+                         rev.parent_ids)
 
         # check next revisions in the history.
         rev = branch.get_revision(self._baz._empty_merged_tag_bzr)
@@ -404,6 +419,39 @@
         self.assertRaises(NoSuchRevision, branch.get_revision, 
                           self._baz._missing_import_bzr)
 
+    def test_missing_ancestor_reusing_history(self):
+        import_version('output', pybaz.Version(self._baz._missing_ancestor),
+                       self.collect,
+                       reuse_history_from=[self._baz._missing_import_imported])
+        # expected results:
+        # one commits, no files, revision identifiers of 
+        # 'demo-gone@DONOTUSE%c--import--0--base-0' and 
+        # 'demo@DONOTUSE%c--gone--0--base-0'
+        branch = Branch.open('output')
+        self.assertEqual(branch.revision_history(),
+                         [self._baz._missing_import_bzr,
+                          self._baz._missing_ancestor_bzr])
+        rev = branch.get_revision(self._baz._missing_ancestor_bzr)
+        # and again.
+        import_version('output2', pybaz.Version(self._baz._missing_ancestor),
+                       self.collect,
+                       reuse_history_from=[self._baz._missing_import_imported])
+        branch2 = Branch.open('output2')
+        self.assertEqual(branch.revision_history(), branch2.revision_history())
+        rev2 = branch2.get_revision(self._baz._missing_ancestor_bzr)
+        # they must be the same
+        self.assertEqual(rev, rev2)
+
+        # must be able to get the missing base revision
+        branch.get_revision(self._baz._missing_import_bzr)
+
+        # and we should get some expected values:
+        self.assertEqual(rev.committer, "Test User<test@example.org>")
+        self.assertEqual(rev.message, "tag of demo-gone@DONOTUSE/c--import--0--base-0")
+        self.assertEqual(rev.revision_id, self._baz._missing_ancestor_bzr)
+        self.assertEqual(rev.parent_ids[0], self._baz._missing_import_bzr)
+        self.assertEqual(1, len(rev.parent_ids))
+
     def test_bad_file_id(self):
         import_version('output', pybaz.Version(self._baz._bad_id_tag),
                        self.collect)
@@ -595,3 +643,13 @@
         command = cmd_baz_import()
         command.run(os.path.join(self._tmpdir, 'output'), 'demo@DONOTUSE')
         command.run(os.path.join(self._tmpdir, 'output'), 'demo@DONOTUSE')
+        
+    def test_accepts_reuse_history(self):
+        self.make_import('c--0')
+        self.run_bzr('baz-import', os.path.join(self._tmpdir, 'output'),
+                     'demo@DONOTUSE', '.', '.')
+        
+    def test_does_not_need_reuse_history(self):
+        self.make_import('c--0')
+        self.run_bzr('baz-import', os.path.join(self._tmpdir, 'output'),
+                     'demo@DONOTUSE')

